# Copyright (C) 2007 Kris Maglione
# See PERMISSIONS

module R9P
	class Fcall <Serial
		module Constants
			MaxWelem = 16
			NoTag	= ~0
			NoFid	= ~0
		end
		include Constants
		Types = []
		field :size, :dword, :size, 4
		field :type, :byte
		field :tag, :word

		def self.type(n)
			Types[n] = self
			@type = n
		end
		def self.from(data)
			type = data[4]
			Types[type].new data
		end

		def initialize(data = nil)
			fields[:type] = self.class.module_eval { @type }
			super(data)
		end

		def succ(args = {})
			Types[type+1].new args.merge(:tag => tag)
		end

		class Tversion <Fcall
			type 100
			field :msize, :dword
			field :version, :word, :string
		end
		class Rversion <Fcall
			type 101
			field :msize, :dword
			field :version, :word, :string
		end

		class Tauth <Fcall
			type 102
			field :afid, :dword
			field :uname, :word, :string
			field :aname, :word, :string
		end
		class Rauth <Fcall
			type 103
			field :aqid, Qid
		end

		class Tattach <Fcall
			type 104
			field :fid, :dword
			field :afid, :dword
			field :uname, :word, :string
			field :aname, :word, :string
		end
		class Rattach <Fcall
			type 105
			field :qid, Qid
		end

		class Terror <Fcall
			type 106
			def initialize
				throw "Illegal 9P tag 'Terror' encountered"
			end
		end
		class Rerror <Fcall
			type 107
			field :ename, :word, :string
		end

		class Tflush <Fcall
			type 108
			field :oldtag, :word
		end
		class Rflush <Fcall
			type 109
		end

		class Twalk <Fcall
			type 110
			field :fid, :dword
			field :newfid, :dword
			field :wname, :word, [ :word, :string ]
		end
		class Rwalk <Fcall
			type 111
			field :wqid, :word, [ Qid ]
		end

		class Topen <Fcall
			type 112
			field :fid, :dword
			field :mode, :byte
		end
		class Ropen <Fcall
			type 113
			field :qid, Qid
			field :iounit, :dword
		end

		class Tcreate <Fcall
			type 114
			field :fid, :dword
			field :name, :word, :string
			field :perm, :dword
			field :mode, :byte
		end
		class Rcreate <Fcall
			type 115
			field :qid, Qid
			field :iounit, :dword
		end

		class Tread <Fcall
			type 116
			field :fid, :dword
			field :offset, :qword
			field :count, :dword
		end
		class Rread <Fcall
			type 117
			field :data, :dword, :string
		end

		class Twrite <Fcall
			type 118
			field :fid, :dword
			field :offset, :qword
			field :data, :dword, :string
		end
		class Rwrite <Fcall
			type 119
			field :count, :dword
		end

		class Tclunk <Fcall
			type 120
			field :fid, :dword
		end
		class Rclunk <Fcall
			type 121
		end

		class Tremove <Tclunk
			type 122
		end
		class Rremove <Fcall
			type 123
		end

		class Tstat <Tclunk
			type 124
		end
		class Rstat <Fcall
			type 125
			field :sstat, :word, :size, 2
			field :stat, Stat
		end

		class Twstat <Rstat
			type 126
		end
		class Rwstat <Fcall
			type 127
		end
	end
end
