# Copyright (C) 2007 Kris Maglione
# See PERMISSIONS

module R9P
	class Serial
		Conv = {
			:byte => [ 1, 'C' ],
			:word => [ 2, 'v' ],
			:dword => [ 4, 'V' ],
		}

		attr_reader :fields, :serial, :opstack

		def initialize(data)
			@fields ||= {}
			@serial ||= ''
			@opstack = []
			print "#{self.class.inspect}.new(#{data.inspect})\n" if $debug > 5
			case data
			when String
				deserialize data
			when Hash
				fields.merge! data
			end
			print "\tFields: #{fields.inspect}\n\n" if $debug > 5
		end

		def deserialize(data = @serial)
			@opstack = []
			@serial = data
			unstack self.class.field_stack
			self
		end
		def serialize
			print "serialize #{self.inspect}\n" if $debug > 4
			@opstack = []
			@serial = stack self.class.field_stack
		end

		def fields
			@fields ||= {}
		end
		def opstack
			@opstack
		end
		def serial
			@serial ||= ""
		end

		def slice
			if @offset > 0
				@serial.slice!(0..@offset-1)
				@offset = 0
			end
		end
		def unstack(stack)
			print "\nunstack(stack): #{self.class.inspect}\n" if $debug > 4
			@offset = 0
			stack.each { |op|
				print "\top: #{op.inspect} top: #{@opstack[-1].inspect}\n" if $debug > 4
				print "\tserial: #{@serial.inspect}\n" if $debug > 6
				case op
				when Field
					fields[op.name] = pop
				when Array
					list = (1..pop).to_a.collect {
						unstack op
						pop
					}
					push list
				when :string
					num = pop
					str, = unpack num, "a#{num}"
					push str
				when :qword
					low, high = unpack 8, "VV"
					push high<<32|low
				when :size
				when Symbol
					push unpack(*Conv[op])[0]
				when Module
					if op <= Serial
						slice
						push op.new(@serial)
					end
				end
			}
			print "\n/unstack(stack): #{self.class.inspect}\n" if $debug > 4
			slice
			self
		end
		def stack(stack)
			print "\nstack(stack): #{self.class.inspect}\n" if $debug > 4
			stack.reverse.each { |op|
				print "\top: #{op.inspect} top: #{@opstack[-1].inspect}\n" if $debug > 4
				print "\tstack: #{@opstack.inspect}\n" if $debug > 5
				print "\tserial: #{@serial.inspect}\n" if $debug > 5
				case op
				when Field
					push @fields[op.name]
				when Array
					list = pop
					list.reverse.each { |i|
						push i
						stack op
					}
					push list.length
				when :string
					str = pop
					push str
					push str.length
				when :qword
					num = pop
					push [num, num >> 32].pack("VV")
				when :size
					add = pop
					push @opstack.inject(add) { |l, o| l + o.length }
				when Symbol
					push [pop].pack(Conv[op][1])
				when Module
					if op <= Serial
						push op.new(@serial).serialize
					end
				else
					pop
					push op
				end
			}
			print "\n/stack(stack): #{self.class.inspect}\n" if $debug > 4
			@opstack.reverse.join ''
		end

		def push(arg)
			@opstack.push arg
		end
		def pop
			@opstack.pop
		end
		def unpack(length, fmt)
			print "unpack(#{length}, #{fmt}) @offset: #{@offset}\n" if $debug > 5
			@offset += length
			@serial.unpack("@#{@offset-length}#{fmt}")
		end

		def to_s
			"fields: #{@fields.inspect}, serial: #{@serial.inspect}, stack: #{@opstack.inspect}"
		end
		def inspect
			"#<#{self.class.inspect} #{to_s}>"
		end

		class Field
			attr_reader :name
			def initialize(name)
				@name = name
			end
		end

		class <<self
			def field_stack
				@field_stack ||=if self.superclass <= Serial
							self.superclass.field_stack.clone
						else
							[]
						end
			end
			def field(name, *rest)
				module_eval <<-"end_eval"
					def #{name.id2name}
						@fields[#{name.inspect}]
					end
					def #{name.id2name}=(val)
						@fields[#{name.inspect}] = val
					end
				end_eval
				self.field_stack.concat(rest) << Field.new(name)
			end
		end
	end
end
