# Copyright (C) 2007 Kris Maglione
# See PERMISSIONS

begin
	require 'thread'
rescue
	# Threading is not necessary for most functions
	module Stub
		class <<self
			def nops(*arg)
				arg.each { |s|
					module_eval <<-"end_eval"
						alias #{s.id2name} nop
					end_eval
				}
			end
			def error(name, *arg)
				first = arg.shift
				module_eval <<-"end_eval"
					def #{first.id2name}(*arg)
						throw #{name.inspect}
					end
				end_eval
				arg.each { |s| module_eval "alias #{s.id2name} #{first.id2name}" }
			end
		end
		def nop(*arg)
			puts "In nop!"
			if block_given?
				yield
			end
		end
	end
	class Thread
		include Stub
		@@dummy = new
		def current
			@@dummy
		end
		Stub.error 'Threading not implemented', :initialize, :wakeup
		Stub.nops :exclusive, :critical, :critical=

		class <<self
			Stub.error 'Threading not implemented', :fork, :start, :stop
		end
	end

	# Locking is not necessary without threading
	class Mutex
		include Stub
		Stub.nops :lock, :unlock, :synchronize, :initialize, :locked?
		def locked?
			false
		end
	end
end

module R9P
	module Qlock
		def qlinit
			@ql_queue = []
			@ql_locked = nil
		end

		def qlock
			cr = Thread.critical
			while (Thread.critical = true; @ql_locked != nil && @ql_locked != Thread.current)
				unless @ql_queue.include?(Thread.current)
					@ql_queue.push Thread.current
				end
				Thread.stop
			end
			@ql_locked = Thread.current
			Thread.critical = cr
			self
		end
		def qunlock
			if @ql_locked != Thread.current
				raise
			end
			cr = Thread.critical
			Thread.critical = true
			thr = @ql_locked = @ql_queue.shift
			@ql_locked.wakeup if @ql_locked
			if block_given?
				yield
			end
		ensure
			Thread.critical = cr
			thr.run if thr
		end
		def qsync
			qlock
			yield
		ensure
			qunlock
		end
		def qlocked?
			@ql_locked ? true : false
		end
	end
	module Rendez
		def sleep
			@thread = Thread.current
			@lock.qunlock {
				Thread.stop
			}
			@lock.qlock
			@thread = nil if @thread == Thread.current
		end
		def wakeup
			if @thread
				@thread.wakeup
			end
		end
	end
end
